/*
 * Copyright (C) 2023 Dominik Drexler and Simon Stahlberg
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef MIMIR_LANGUAGES_GENERAL_POLICIES_FORMATTER_HPP_
#define MIMIR_LANGUAGES_GENERAL_POLICIES_FORMATTER_HPP_

#include "mimir/languages/general_policies/declarations.hpp"

namespace mimir
{
namespace graphs
{
extern std::ostream& operator<<(std::ostream& out, const PolicyVertex& element);

extern std::ostream& operator<<(std::ostream& out, const PolicyEdge& element);
}

namespace languages::general_policies
{
extern std::ostream& operator<<(std::ostream& out, const ICondition& element);

extern std::ostream& operator<<(std::ostream& out, const IEffect& element);

template<dl::IsConceptOrRoleOrBooleanOrNumericalTag D>
std::ostream& operator<<(std::ostream& out, const NamedFeatureImpl<D>& element);

extern std::ostream& operator<<(std::ostream& out, const RuleImpl& element);

extern std::ostream& operator<<(std::ostream& out, const GeneralPolicyImpl& element);
}  // end languages::general_policies

extern std::ostream& print(std::ostream& out, const mimir::graphs::PolicyVertex& element);

extern std::ostream& print(std::ostream& out, const mimir::graphs::PolicyEdge& element);

extern std::ostream& print(std::ostream& out, const mimir::languages::general_policies::ICondition& element);

extern std::ostream& print(std::ostream& out, const mimir::languages::general_policies::IEffect& element);

template<mimir::languages::dl::IsConceptOrRoleOrBooleanOrNumericalTag D>
std::ostream& print(std::ostream& out, const mimir::languages::general_policies::NamedFeatureImpl<D>& element);

extern std::ostream& print(std::ostream& out, const mimir::languages::general_policies::RuleImpl& element);

extern std::ostream& print(std::ostream& out, const mimir::languages::general_policies::GeneralPolicyImpl& element);
}

#endif
